from typing import Optional
from typing import Optional
from PIL import Image
from ..codec import ImageCodecAdaptor, supportsFormats
from ..constants import CompressionFormat, CompressionQuality
from ..exceptions import DependancyMissingException

try:
    import tex2img
except ImportError as e:
    raise DependancyMissingException("Tex2ImgCodec", "tex2img", e)

TEX2IMG_FORMAT_MAP = {
    # CompressionFormat.PVRTC14A: 12, # Tex2ImgCodec segfaults decoding PVRTC with all tests (#29)
    CompressionFormat.ATC: 14,
    CompressionFormat.DXT1: 5,
    CompressionFormat.DXT5: 6,
    CompressionFormat.ETC1: 0,
    CompressionFormat.ETC2: 2
}


@supportsFormats(decompresses=TEX2IMG_FORMAT_MAP.keys())
class Tex2ImgCodec(ImageCodecAdaptor):
    @classmethod
    def decompress(cls, fp: bytes, format: CompressionFormat, width: int, height: int, quality: Optional[CompressionQuality]) -> Image.Image:
        if format not in TEX2IMG_FORMAT_MAP:
            raise ValueError(f"Codec {Tex2ImgCodec.__name__} does not support format {format.name}")
        
        decompressed = tex2img.basisu_decompress(fp, width, height, TEX2IMG_FORMAT_MAP[format]) # type: ignore[reportUnknownMemberType]

        return Image.frombytes("RGBA", (width, height), decompressed, "raw") # type: ignore[reportUnknownMemberType]



import struct
from typing import Literal, NamedTuple, BinaryIO, Optional, TypeVar, Union

ShortEndianness = Literal["<", ">"]
NameEndianness = Literal["little", "big"]

class Endianness(NamedTuple):
    name: NameEndianness
    short: ShortEndianness


def uint8(x: int, endianness: Endianness) -> bytes:
    """Convert an integer to a C unsigned char.

    :param x: The integer to convert
    :type x: int
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: `x` in C uint8-representation
    :rtype: bytes
    """
    return struct.pack(endianness.short + "B", x)


def uint16(x: int, endianness: Endianness) -> bytes:
    """Convert an integer to a C unsigned short.

    :param x: The integer to convert
    :type x: int
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: `x` in C uint16-representation
    :rtype: bytes
    """
    return struct.pack(endianness.short + "H", x)

    
def uint32(x: int, endianness: Endianness) -> bytes:
    """Convert an integer to a C unsigned int.

    :param x: The integer to convert
    :type x: int
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: `x` in C uint32-representation
    :rtype: bytes
    """
    return struct.pack(endianness.short + "I", x)


TDefault = TypeVar("TDefault", bound=Optional[int])

def intFromBytes(b: bytes, endianness: Endianness, default: TDefault = 0) -> Union[int, TDefault]:
    """Interpret `b` as a python integer.

    :param b: The bytes to interpret
    :type b: bytes
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: `b` interpreted as a python int
    :rtype: bytes
    """
    if b == b'':
        return default
    return int.from_bytes(b, byteorder=endianness.name)


def readUInt8(fp: BinaryIO, endianness: Endianness, default: TDefault = 0) -> Union[int, TDefault]:
    """Read a C unsigned char into a python integer.

    :param x: The binary stream from which to read
    :type x: BinaryIO
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: The next 1 byte from `fp` interpreted as a C uint8
    :rtype: bytes
    """
    return intFromBytes(fp.read(1), endianness, default)


def readUInt16(fp: BinaryIO, endianness: Endianness, default: TDefault = 0) -> Union[int, TDefault]:
    """Read a C unsigned short into a python integer.

    :param x: The binary stream from which to read
    :type x: BinaryIO
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: The next 2 bytes from `fp` interpreted as a C uint16
    :rtype: bytes
    """
    return intFromBytes(fp.read(2), endianness, default)

    
def readUInt32(fp: BinaryIO, endianness: Endianness, default: TDefault = 0) -> Union[int, TDefault]:
    """Read a C unsigned int into a python integer.

    :param x: The binary stream from which to read
    :type x: BinaryIO
    :param endianness: The bit-endianness
    :type endianness: Endianness
    :return: The next 4 bytes from `fp` interpreted as a C uint32
    :rtype: bytes
    """
    return intFromBytes(fp.read(4), endianness, default)



from typing import Optional
from . import constants

def _append(base: str, extra: Optional[str]):
    return base + (f" {extra}" if extra else "")


class AEPiException(Exception):
    """Base class for all AEPi exceptions.
    """
    def __init__(self, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(*args)
        self.message = message
        self.inner = inner


    def __str__(self) -> str:
        items = [str(i) for i in (self.message, self.inner) if i]
        return "\n".join(items) + super().__str__()


class CodecLoadException(AEPiException):
    """Thrown when a codec failed to initialize, and so should be ignored.
    
    :var str codecName: The name of the codec which failed to load
    """
    def __init__(self, codecName: str, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(_append(f"Codec {codecName} failed to initialize.", message), inner, *args)
        self.codecName = codecName


class DependencyFailedException(CodecLoadException):
    """Thrown when a codec failed to initialize due to a dependency error.

    :var str dependencyName: The name of the dependency which failed
    :var Optional[Exception] inner: The exception which caused the failure, if applicable
    """
    def __init__(self, codecName: str, dependencyName: str, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(codecName, _append(f"Dependency '{dependencyName}' failed to load.", message), inner, *args)
        self.dependencyName = dependencyName


class DependancyMissingException(DependencyFailedException):
    """Thrown when a codec failed to initialize because a dependency is not installed.
    """
    def __init__(self, codecName: str, dependencyName: str, inner: ImportError, message: Optional[str] = None, *args: object) -> None:
        super().__init__(codecName, dependencyName, _append("The dependency is not installed.", message), inner, *args)


class AeiReadException(AEPiException):
    """Base class for exceptions thrown whilst reading an existing AEI file.
    """
    def __init__(self, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(_append("The AEI could not be read.", message), inner, *args)


class AeiWriteException(AEPiException):
    """Base class for exceptions thrown whilst writing a new AEI file.
    """
    def __init__(self, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(_append("The AEI could not be written.", message), inner, *args)


class InvalidCompressionFormatException(AeiReadException):
    """Thrown when an AEI file specified an unsupported compression format.
    """
    def __init__(self, formatId: int, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(_append(f"The AEI specified an unknown compression format: {formatId}.", message), inner, *args)
        self.compressionFormatId = formatId


class UnsupportedAeiFeatureException(AEPiException):
    """Thrown when an AEI contains a feature that AEPi does not yet support
    """
    def __init__(self, featureName: str, message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(_append(f"This operation requires a feature that AEPi does not yet support: {featureName}", message), inner, *args)


class UnsupportedCompressionFormatException(UnsupportedAeiFeatureException):
    """Thrown when an AEI is compressed in a valid format that AEPi does not yet support
    """
    def __init__(self, format: "constants.CompressionFormat", message: Optional[str] = None, inner: Optional[Exception] = None, *args: object) -> None:
        super().__init__(f"Compression format '{format.name}'", message, inner, *args)
        self.compressionFormat = format



import importlib
from ..exceptions import CodecLoadException

# Discover codecs

_CODECS = ["EtcPakCodec", "Tex2ImgCodec", "RawCodec"]

for codec in _CODECS:
    try:
        importlib.import_module(f".{codec}", "AEPi.codecs")
    except CodecLoadException:
        pass



import io
from typing import Any, BinaryIO
from os import PathLike
from types import TracebackType
from typing import Any, List, Optional, Set, Tuple, Type, TypeVar, Union, cast, overload
from PIL import Image
from contextlib import nullcontext

from ..lib import imageOps
from ..lib.binaryio import uint8, uint16, uint32, readUInt8, readUInt16, readUInt32

from ..constants import CompressionFormat, FILE_TYPE_HEADER, ENDIANNESS, CompressionQuality
from .. import codec
from .texture import Texture
from ..exceptions import UnsupportedAeiFeatureException, AeiReadException, AeiWriteException

TException = TypeVar("TException", bound=Exception)

class AEI:
    """An Abyss Engine Image file.
    Contains a set of textures, each with an image and coordinates.
    Each texture must fall within the bounds of the AEI shape.
    The AEI shape is mutable, through the `shape` property.
    The coordinate origin (0, 0) is the top-left of the AEI.

    An AEI can be constructed either with its dimensions, or with an image.
    If an image is used, the AEI is created with a copy of the image.
    `format` and `quality` can be set in the constructor, or on call of `AEI.write`.

    Use the `addTexture` and `removeTexture` helper methods for texture management.

    To decode an existing AEI file, use `AEI.read`.
    To encode an AEI into a file, use `AEI.write`.

    If the AEI is scoped in a `with` statement, when exiting the `with`,
    the AEI will attempt to close all images, and swallow any errors encountered.
    """
    @overload
    def __init__(self, shape: Tuple[int, int], /, format: Optional[CompressionFormat] = None, quality: Optional[CompressionQuality] = None) -> None: ...
    
    @overload
    def __init__(self, image: Image.Image, /, format: Optional[CompressionFormat] = None, quality: Optional[CompressionQuality] = None) -> None: ...

    def __init__(self, val1: Union[Image.Image, Tuple[int, int]], /, format: Optional[CompressionFormat] = None, quality: Optional[CompressionQuality] = None):
        self._textures: List[Texture] = []
        self._texturesWithoutImages: Set[Texture] = set()
        self.format = format
        self.quality: Optional[CompressionQuality] = quality

        if isinstance(val1, Image.Image):
            self._shape = val1.size
            self._image = val1.copy()
        else:
            self._shape = val1
            self._image = Image.new("RGBA", self._shape)


    @property
    def shape(self):
        """The dimensions of the AEI, in pixels.

        :return: (width, height)
        :rtype: Tuple[int, int]
        """
        return self._shape
    

    @shape.setter
    def shape(self, value: Tuple[int, int]):
        widthShrunk = value[0] < self.shape[0]
        heightShrunk = value[1] < self.shape[1]

        if widthShrunk or heightShrunk:
            for tex in self.textures:
                if widthShrunk and tex.x + tex.width > value[0] \
                        or heightShrunk and tex.y + tex.height > value[1]:
                    raise ValueError(f"Changing shape from ({self.shape[0]}, {self.shape[1]}) to ({value[0]}, {value[1]}) would cause texture ({tex.x}, {tex.y}) to fall out of bounds")
        
        self._shape = value


    @property
    def width(self):
        return self.shape[0]
    

    @width.setter
    def width(self, value: int):
        self.shape = (value, self.height)
    

    @property
    def height(self):
        return self.shape[1]
    

    @height.setter
    def height(self, value: int):
        self.shape = (self.width, value)


    @property
    def textures(self):
        """Do not use this property to manage textures. Instead use `addTexture` and `removeTexture`.
        This is the mutable, internal representation. Altering directly could cause issues.

        :return: The textures within the AEI
        :rtype: List[Texture]
        """
        return self._textures
    

    def _validateBoundingBox(self, val1: Union[Texture, int], y: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None) -> Tuple[int, int, int, int]:
        if isinstance(val1, Texture):
            y = val1.y
            width = val1.width
            height = val1.height
            val1 = val1.x
            
        elif y is None or width is None or height is None:
            raise ValueError("All of x, y, width and height are required")
        
        if val1 < 0 or width < 1 or y < 0 or height < 1 or val1 + width > self.width or y + height > self.height:
            raise ValueError("The bounding box falls out of bounds of the AEI")
        
        return (val1, y, width, height)


    def _findTextureByBox(self, val1: Union[Texture, int], y: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None):
        x, y, width, height = self._validateBoundingBox(val1, y, width, height)
        
        try:
            texture = next(t for t in self.textures if t.x == x and t.y == y and t.width == width and t.height == height)
        except StopIteration:
            return None
        
        return texture


    @overload
    def addTexture(self, texture: Texture, /) -> None:
        """Add only a texture bounding box to this AEI. Use the overload with `Image` as the first parameter to add image content.
        This overload is only really useful for creating overlapping textures.

        :param texture: The new texture
        :type texture: Texture
        :raises ValueError: If the `texture` falls out of bounds of the AEI
        """
    
    @overload
    def addTexture(self, x: int, y: int, width: int, height: int, /) -> None:
        """Add only a texture bounding box to this AEI. Use the overload with `Image` as the first parameter to add image content.
        This overload is only really useful for creating overlapping textures.

        :param int x: The x-coordinate of the texture
        :param int y: The y-coordinate of the texture
        :param int width: The width of the texture
        :param int height: The height of the texture
        :raises ValueError: If the the bounding box falls out of bounds of the AEI
        """

    @overload
    def addTexture(self, image: Image.Image, x: int, y: int, /) -> None:
        """Add a texture with image content to this AEI.
        `image` is not retained, and can be closed after passing to this method without side effects.

        :param image: The new image
        :type image: Image.Image
        :param int x: The x-coordinate of the texture
        :param int y: The y-coordinate of the texture
        :raises ValueError: If `image.mode` is not `RGBA`
        :raises ValueError: If the bounding box falls out of bounds of the AEI
        """

    def addTexture(self, val1: Union[Image.Image, Texture, int], val2: Optional[int] = None, val3: Optional[int] = None, val4: Optional[int] = None, /):
        if isinstance(val1, Texture):
            image = None
            texture = val1
        elif isinstance(val1, int):
            if val2 is None or val3 is None or val4 is None:
                raise ValueError("All of x, y, width and height are required")
            image = None
            texture = Texture(val1, val2, val3, val4)
        elif val2 is None or val3 is None:
            raise ValueError("Both x and y are required")
        else:
            image = val1
            texture = Texture(val2, val3, image.width, image.height) 

        existingTexture = self._findTextureByBox(texture)
        if existingTexture is not None:
            raise ValueError("A texture already exists with the given bounding box")
        
        if image is None:
            self._texturesWithoutImages.add(texture)
        else:
            if texture.width != image.width or texture.height != image.height:
                raise ValueError("image dimensions do not match the texture dimensions")

            if image.mode != "RGBA":
                raise ValueError(f"image must be mode RGBA, but {image.mode} was given")
            
            self._image.paste(image, (texture.x, texture.y), image)
        
        self.textures.append(texture)


    def replaceTexture(self, image: Image.Image, texture: Texture):
        """Replace a texture in this AEI.
        `image` is not retained, and can be closed after passing to this method without side effects.

        :param image: The new image
        :type image: Image.Image
        :param texture: The new texture
        :type texture: Texture
        :raises ValueError: If `image.mode` is not `RGBA`
        :raises ValueError: If the `texture` falls out of bounds of the AEI
        :raises ValueError: If the dimensions in `texture` do not match the dimensions of `image`
        """
        existingTexture = self._findTextureByBox(texture)
        if existingTexture is None:
            raise KeyError(f"no texture was found with coordinates ({texture.x}, {texture.y}) and dimensions ({texture.width}, {texture.height})")
        
        if texture.width != image.width or texture.height != image.height:
            raise ValueError("image dimensions do not match the texture dimensions")

        if image.mode != "RGBA":
            raise ValueError(f"image must be mode RGBA, but {image.mode} was given")
        
        self._image.paste(image, (texture.x, texture.y), image)


    @overload
    def removeTexture(self, texture: Texture, /, *, clearImage: Optional[bool] = None) -> None: ...

    @overload
    def removeTexture(self, x: int, y: int, width: int, height: int, /, *, clearImage: Optional[bool] = None) -> None: ...

    def removeTexture(self, val1: Union[Texture, int], y: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None, /, *, clearImage: Optional[bool] = None):
        """Remove a texture from this AEI, by its bounding box.

        :param Optional[bool] clearImage: clear the area of the image. Default: Clear if an image was provided when the texture was added
        :raises KeyError: If no corresponding texture could be found for the bounding box
        """
        texture = self._findTextureByBox(val1, y, width, height)
        if texture is None:
            raise KeyError(f"no texture was found with coordinates ({val1}, {y}) and dimensions ({width}, {height})")
        
        if clearImage is not None and clearImage or clearImage is None and texture not in self._texturesWithoutImages:
            # Clear the area that the texture occupied
            self._image.paste(
                (0, 0, 0, 0),
                (texture.x, texture.y, texture.x + texture.width, texture.y + texture.height)
            )

        self._textures.remove(texture)
    

    @overload
    def getTexture(self, texture: Texture, /) -> Image.Image: ...

    @overload
    def getTexture(self, x: int, y: int, width: int, height: int, /) -> Image.Image: ...

    def getTexture(self, val1: Union[Texture, int], y: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None, /) -> Image.Image:
        """Get a copy of the image defined by the provided bounding box.

        :returns: a copy of the image defined by the provided bounding box
        :rtype: Image.Image
        :raises KeyError: The provided bounding box falls out of bounds of the AEI
        """
        x, y, width, height = self._validateBoundingBox(val1, y, width, height)
        
        return self._image.crop((x, y, x + width, y + height))


    @classmethod
    def read(cls, fp: Union[str, PathLike[Any], io.BytesIO]) -> "AEI":
        """Read an AEI file from bytes, or a file.
        `fp` can be a path to a file, or an in-memory buffer containing the contents of an encoded AEI file, including metadata.

        :param fp: The AEI itself, or a path to an AEI file on disk
        :type fp: Union[str, PathLike, io.BytesIO]
        :return: A new AEI file object, containing the decoded contents of `fp`
        :rtype: AEI
        """
        if isinstance(fp, io.StringIO):
            raise ValueError("fp must be of binary type, not StringIO")
        
        file: Union[io.BufferedReader, io.BytesIO]

        if tempFp := (not isinstance(fp, io.BytesIO)):
            file = open(fp, "rb")
        elif isinstance(fp, (str, PathLike)):
            file = open(fp, mode="rb")
        else:
            file = fp

        try:
            bFileType = file.read(len(FILE_TYPE_HEADER))
            if bFileType != FILE_TYPE_HEADER:
                raise ValueError(f"Given file is of unknown type '{str(bFileType, encoding='utf-8')}' expected '{str(FILE_TYPE_HEADER, encoding='utf-8')}'")

            formatId = readUInt8(file, ENDIANNESS)
            format, mipmapped = CompressionFormat.fromBinary(formatId)
            # if mipmapped:
            #     raise UnsupportedAeiFeatureException("Mipmapped textures")
            
            imageCodec = codec.decompressorFor(format)

            width = readUInt16(file, ENDIANNESS)
            height = readUInt16(file, ENDIANNESS)
            numTextures = readUInt16(file, ENDIANNESS)

            textures: List[Texture] = []
            for _ in range(numTextures):
                texX = readUInt16(file, ENDIANNESS)
                texY = readUInt16(file, ENDIANNESS)
                texWidth = readUInt16(file, ENDIANNESS)
                texHeight = readUInt16(file, ENDIANNESS)
                textures.append(Texture(texX, texY, texWidth, texHeight))

            if format.isCompressed:
                imageLength = readUInt32(file, ENDIANNESS)
            else:
                imageLength = 4 * width * height

            compressed = file.read(imageLength)

            symbolGroups = readUInt16(file, ENDIANNESS)

            if symbolGroups > 0:
                raise UnsupportedAeiFeatureException("Symbol maps")
            
            bQuality = readUInt8(file, ENDIANNESS, None)
            quality = cast(Optional[CompressionQuality], bQuality) 

            decompressed = imageCodec.decompress(compressed, format, width, height, quality)

            if format.isBgra:
                with decompressed:
                    imageContent = imageOps.switchRGBA_BGRA(decompressed)
            else:
                imageContent = decompressed

            if imageContent.mode != format.pillowMode:
                with imageContent:
                    imageContent = imageContent.convert(format.pillowMode)

        except Exception as ex:
            raise AeiReadException(None, ex) from ex

        finally:
            if tempFp:
                file.close()
        
        aei = AEI(imageContent, format=format, quality=quality)
        for tex in textures:
            aei.addTexture(tex)

        return aei
    

    def write(self, fp: Optional[BinaryIO] = None, format: Optional[CompressionFormat] = None, quality: Optional[CompressionQuality] = None) -> BinaryIO:
        """Write this AEI to a BytesIO file.

        :param fp: Optional file to write to. If not given, a new one is created. defaults to None
        :type fp: Optional[io.BytesIO], optional
        :param format: Override for the compression format. defaults to the setting on the AEI
        :type format: Optional[CompressionFormat], optional
        :param quality: Override for the compression quality. defaults to the setting on the AEI
        :type quality: Optional[CompressionQuality], optional
        :raises ValueError: If format is omitted and no format is set on the AEI
        :raises ValueError: If the AEI has no textures
        :return: A file containing the AEI, including the compressed image and full metadata
        :rtype: io.BytesIO
        """
        format = format or self.format
        quality = self.quality if quality is None else quality

        if format is None:
            raise ValueError("This AEI has no compression format specified. Set self.format, or specify the format in the self.toFile.format kwarg")

        fp = io.BytesIO() if fp is None else fp

        # AEIs must contain at least one texture
        if tempTexture := (len(self.textures) == 0):
            self.addTexture(Texture(0, 0, self.width, self.height))

        try:
            self._writeHeaderMeta(fp, format)
            self._writeImageContent(fp, format, quality)
            self._writeSymbols(fp)
            self._writeFooterMeta(fp, quality)

        except Exception as ex:
            raise AeiWriteException(None, ex) from ex
        
        finally:
            if tempTexture:
                self.removeTexture(0, 0, self.width, self.height)

        return fp
    
#region write-util
    
    def _writeHeaderMeta(self, fp: BinaryIO, format: CompressionFormat):
        fp.write(FILE_TYPE_HEADER)
        fp.write(uint8(format.value, ENDIANNESS))

        def writeUInt16(*values: int):
            for v in values:
                fp.write(uint16(v, ENDIANNESS))

        # AEI dimensions and texture count
        writeUInt16(
            self.width,
            self.height,
            len(self.textures)
        )

        # texture bounding boxes
        for texture in self.textures:
            writeUInt16(
                texture.x,
                texture.y,
                texture.width,
                texture.height
            )
    

    def _writeImageContent(self, fp: BinaryIO, format: CompressionFormat, quality: Optional[CompressionQuality]):
        imageCodec = codec.compressorFor(format)

        if format.isBgra:
            imageContent = imageOps.switchRGBA_BGRA(self._image)
            ctx = imageContent
        else:
            imageContent = self._image
            ctx = nullcontext()

        with ctx:
            if imageContent.mode != format.pillowMode:
                imageContent = imageContent.convert(format.pillowMode)
                ctx = imageContent
            else:
                ctx = nullcontext()

            with ctx:
                compressed = imageCodec.compress(imageContent, format, quality)

        # image length only appears in compressed AEIs
        if format.isCompressed:
            fp.write(uint32(len(compressed), ENDIANNESS))

        fp.write(compressed)


    def _writeSymbols(self, fp: BinaryIO):
        #TODO: Unimplemented
        fp.write(uint16(0, ENDIANNESS)) # number of symbol groups
        ...


    def _writeFooterMeta(self, fp: BinaryIO, quality: Optional[CompressionQuality]):
        if quality is not None:
            fp.write(uint8(quality, ENDIANNESS))

#endregion write-util

    def close(self):
        """Close the underlying image.
        """
        self._image.close()


    def __enter__(self):
        """This method is called when entering a `with` statement.
        """
        return self


    def __exit__(self, exceptionType: Type[TException], exception: TException, trace: TracebackType):
        """This method is called when exiting a `with` statement.
        """
        try:
            self.close()
        except:
            pass



from typing import Optional
from typing import Optional
from PIL import Image
from ..codec import ImageCodecAdaptor, supportsFormats
from ..constants import CompressionFormat, CompressionQuality

@supportsFormats(both=[
    CompressionFormat.Uncompressed_UI,
    CompressionFormat.Uncompressed_CubeMap_PC,
    CompressionFormat.Uncompressed_CubeMap
])
class RawCodec(ImageCodecAdaptor):
    @classmethod
    def compress(cls, im: Image.Image, format: CompressionFormat, quality: Optional[CompressionQuality]) -> bytes:
        return im.tobytes() # type: ignore[reportUnknownMemberType]
    
    
    @classmethod
    def decompress(cls, fp: bytes, format: CompressionFormat, width: int, height: int, quality: Optional[CompressionQuality]) -> Image.Image:
        return Image.frombytes("RGBA", (width, height), fp, "raw") # type: ignore[reportUnknownMemberType]
    


from abc import ABC
from typing import Dict, Optional, Type, TypeVar, Iterable
from PIL.Image import Image

from .constants import CompressionFormat, CompressionQuality
from .exceptions import UnsupportedCompressionFormatException

class ImageCodecAdaptor(ABC):
    @classmethod
    def compress(cls, im: Image, format: CompressionFormat, quality: Optional[CompressionQuality]) -> bytes:
        """Compress a BGRA image into format `format`, with quality `quality`

        :param im: The image to compress
        :type im: Image
        :param format: The compression format
        :type format: CompressionFormat
        :param quality: The compression quality
        :type quality: CompressionQuality
        :return: `im`, compressed into format `format`
        :rtype: bytes
        """
        raise NotImplementedError(f"Codec {cls.__name__} is not capable of compression")


    @classmethod
    def decompress(cls, fp: bytes, format: CompressionFormat, width: int, height: int, quality: Optional[CompressionQuality]) -> Image:
        """Decompress a `format`-compressed BGRA image into a BGRA Image.

        :param fp: The compressed image to decompress
        :type im: bytes
        :param format: The compression format
        :type format: CompressionFormat
        :param width: The width of the image
        :type width: int
        :param height: The height of the image
        :type height: int
        :return: `fp`, decompressed into a BGRA image
        :rtype: Image
        """
        raise NotImplementedError(f"Codec {cls.__name__} is not capable of decompression")


compressors: Dict[CompressionFormat, Type[ImageCodecAdaptor]] = {}
decompressors: Dict[CompressionFormat, Type[ImageCodecAdaptor]] = {}

TCodec = TypeVar("TCodec", bound=ImageCodecAdaptor)

def supportsFormats(
        compresses: Optional[Iterable[CompressionFormat]] = None,
        decompresses: Optional[Iterable[CompressionFormat]] = None,
        both: Optional[Iterable[CompressionFormat]] = None
    ):
    """Class decorator marking an image codec as able to compress/decompress images into the given compression formats.
    The codec class must assume that BGRA is passed, and return BGRA.

    ```py
    supportsFormats(
        compresses=[format1],
        decompresses=[format1]
    )
    ```

    is equivalent to:

    ```py
    supportsFormats(
        both=[format1]
    )
    ```

    :param compresses: The formats that the codec can compress. defaults to None
    :type format: Optional[Iterable[CompressionFormat]]
    :param decompresses: The formats that the codec can decompress. defaults to None
    :type format: Optional[Iterable[CompressionFormat]]
    :param both: The formats that the codec can compress AND decompress. defaults to None
    :type format: Optional[Iterable[CompressionFormat]]
    """
    def inner(cls: Type[TCodec]) -> Type[TCodec]:
        if compresses:
            for f in compresses:
                compressors[f] = cls

        if decompresses:
            for f in decompresses:
                decompressors[f] = cls

        if both:
            for f in both:
                compressors[f] = cls
                decompressors[f] = cls
        return cls
    return inner


def compressorFor(format: CompressionFormat) -> Type[ImageCodecAdaptor]:
    """Get the codec for the given compression format.

    :param format: The compression format for which the codec should support compression
    :type format: CompressionFormat
    :return: An ImageCodecAdaptor subclass that can compress `format`
    :rtype: Type[ImageCodecAdaptor]
    :raises AeiWriteException: If no compatible codec is loaded
    """
    if format not in compressors:
        raise UnsupportedCompressionFormatException(format)
    return compressors[format]


def decompressorFor(format: CompressionFormat) -> Type[ImageCodecAdaptor]:
    """Get the codec for the given decompression format.

    :param format: The compression format for which the codec should support decompression
    :type format: CompressionFormat
    :return: An ImageCodecAdaptor subclass that can decompress `format`
    :rtype: Type[ImageCodecAdaptor]
    :raises AeiReadException: If no compatible codec is loaded
    """
    if format not in decompressors:
        raise UnsupportedCompressionFormatException(format)
    return decompressors[format]



from .image import AEI, Texture
from .constants import CompressionFormat, CompressionQuality
from .codec import *
from . import codecs
from . import lib

__version__ = "0.8.3.1"
__all__ = ["AEI", "Texture", "CompressionFormat", "CompressionQuality", "codecs", "lib", "codec"]



from typing import Optional
from typing import Optional
from PIL.Image import Image
from contextlib import nullcontext
from ..codec import ImageCodecAdaptor, supportsFormats
from ..constants import CompressionFormat, CompressionQuality
from ..exceptions import DependancyMissingException

try:
    import etcpak
except ImportError as e:
    raise DependancyMissingException("EtcPakCodec", "etcpak", e)


@supportsFormats(compresses=[
    CompressionFormat.DXT5,
    CompressionFormat.ETC1
])
class EtcPakCodec(ImageCodecAdaptor):
    @classmethod
    def compress(cls, im: Image, format: CompressionFormat, quality: Optional[CompressionQuality]) -> bytes:
        imageIn, ctx = cls._ensureRgba(im)
        with ctx:
            if format is CompressionFormat.DXT5:
                return etcpak.compress_to_dxt5(imageIn.tobytes(), imageIn.width, imageIn.height) # type: ignore[reportUnknownVariableType]

            elif format is CompressionFormat.ETC1:
                return etcpak.compress_to_etc1(imageIn.tobytes(), imageIn.width, imageIn.height) # type: ignore[reportUnknownVariableType]
            
        raise ValueError(f"Codec {EtcPakCodec.__name__} does not support format {format.name}")
    
    
    @classmethod
    def _ensureRgba(cls, im: Image):
        converted = im.mode != "RGBA"
        if converted:
            rgbaImg = im.convert("RGBA")
        else:
            rgbaImg = im
        
        return rgbaImg, (rgbaImg if converted else nullcontext())



from enum import Enum
from typing import Literal, Dict, Set, Tuple
from .lib.binaryio import Endianness
from . import exceptions 

FORMAT_PILLOW_MODES: Dict["CompressionFormat", str] = {}
FORMAT_BITCOUNTS: Dict["CompressionFormat", int] = {}
BGR_FORMATS: Set["CompressionFormat"] = set()
MIPMAPPABLE_FORMATS: Set["CompressionFormat"] = set()
MASK_MIPMAPPED_FLAG = 0b00000010
MASK_FORMAT_ID = 0b11111101

class CompressionFormat(Enum):
    """Compression formats.
    Enum values correspond to the identifiers used within AEI files.
    """
    Unknown =                   0b00000000
    Uncompressed =              0b00000011
    Uncompressed_UI =           0b00000001
    Uncompressed_CubeMap_PC =   0b10000001
    Uncompressed_CubeMap =      0b11000010
    PVRTC12A =                  0b00001101
    PVRTC14A =                  0b00010000
    ATC =                       0b00010001
    DXT1 =                      0b00100000
    DXT3 =                      0b00100001
    DXT5 =                      0b00100100
    ETC1 =                      0b01000000
    ETC2 =                      0b00010111


    @classmethod
    def fromBinary(cls, rawId: int) -> Tuple["CompressionFormat", bool]:
        # Some non-mipmappable formats have the mipmapping flag as 1, just as part of their ID
        if any(f.value == rawId for f in CompressionFormat):
            return CompressionFormat(rawId), False
        
        mipmapped = rawId & MASK_MIPMAPPED_FLAG == 2
        id = rawId & MASK_FORMAT_ID
        
        if not any(f.value == id for f in CompressionFormat):
            raise exceptions.InvalidCompressionFormatException(rawId)
        
        return CompressionFormat(id), mipmapped


    @property
    def isCompressed(self):
        return self not in (
            CompressionFormat.Unknown,
            CompressionFormat.Uncompressed_UI,
            CompressionFormat.Uncompressed_CubeMap,
            CompressionFormat.Uncompressed_CubeMap_PC
        )
    

    @property
    def pillowMode(self):
        return FORMAT_PILLOW_MODES[self]
    

    @property
    def bitcount(self):
        return FORMAT_BITCOUNTS[self]
    
    
    @property
    def supportsMipmapping(self):
        # This will need some more testing to validate
        return self in MIPMAPPABLE_FORMATS
    
    
    @property
    def isBgra(self):
        return self in BGR_FORMATS


FORMAT_PILLOW_MODES[CompressionFormat.Uncompressed] = "RGBA" # ?
FORMAT_PILLOW_MODES[CompressionFormat.Uncompressed_UI] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.Uncompressed_CubeMap_PC] = "RGBA" # ?
FORMAT_PILLOW_MODES[CompressionFormat.Uncompressed_CubeMap] = "RGBA" # ?
FORMAT_PILLOW_MODES[CompressionFormat.PVRTC12A] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.PVRTC14A] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.ATC] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.DXT1] = "RGB"
FORMAT_PILLOW_MODES[CompressionFormat.DXT3] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.DXT5] = "RGBA"
FORMAT_PILLOW_MODES[CompressionFormat.ETC1] = "RGB"
FORMAT_PILLOW_MODES[CompressionFormat.ETC2] = "RGB"

FORMAT_BITCOUNTS[CompressionFormat.Uncompressed] = 8 # ?
FORMAT_BITCOUNTS[CompressionFormat.Uncompressed_UI] = 8 # ?
FORMAT_BITCOUNTS[CompressionFormat.Uncompressed_CubeMap_PC] = 8 # ?
FORMAT_BITCOUNTS[CompressionFormat.Uncompressed_CubeMap] = 8 # ?
FORMAT_BITCOUNTS[CompressionFormat.PVRTC12A] = 2
FORMAT_BITCOUNTS[CompressionFormat.PVRTC14A] = 4
FORMAT_BITCOUNTS[CompressionFormat.ATC] = 4
FORMAT_BITCOUNTS[CompressionFormat.DXT1] = 4
FORMAT_BITCOUNTS[CompressionFormat.DXT3] = 8
FORMAT_BITCOUNTS[CompressionFormat.DXT5] = 8
FORMAT_BITCOUNTS[CompressionFormat.ETC1] = 4
FORMAT_BITCOUNTS[CompressionFormat.ETC2] = 8 # ?

BGR_FORMATS.add(CompressionFormat.ETC1)

MIPMAPPABLE_FORMATS.add(CompressionFormat.PVRTC12A)
MIPMAPPABLE_FORMATS.add(CompressionFormat.PVRTC14A)
MIPMAPPABLE_FORMATS.add(CompressionFormat.ATC)
MIPMAPPABLE_FORMATS.add(CompressionFormat.DXT1)
MIPMAPPABLE_FORMATS.add(CompressionFormat.DXT3)
MIPMAPPABLE_FORMATS.add(CompressionFormat.DXT5)
MIPMAPPABLE_FORMATS.add(CompressionFormat.ETC1)

FILE_TYPE_HEADER = b"AEimage\x00"
ENDIANNESS = Endianness("little", "<")
CompressionQuality = Literal[1, 2, 3]
